(* :Mathematica Version: Mathematica 4.0 *)

(* :Name: Package`PlotMesh` *)

(* :Context: Package`CreateMesh` *)

(* :Title: Parametric Plots of 2D and 3D Objects in 2D respectively 3D *)

(* :Author:
	Andrei Constantinescu 
*)

(* :Copyright: Copyright 2006, Andrei Constantinescu, CNRS, LMS, Ecole Polytechnique France *)

(* :Package Version: 1.0 *)

(* :History:
    V1.0 by Andrei Constantinescu
    
This notebook forms part of the supporting material for the book

Elasticity with Mathematica (r)
by A. Constantinescu and A.M. Korsunsky
published by Cambridge University Press, 2007
(ISBN-13: 9780521842013)

Copyright (c) 2007 A. Constantinescu and A.M. Korsunsky

Please acknowledge the use of this material in teaching, research and 
further development.

Authors accept no responsibility for the results obtained and 
conclusions drawn on the basis of using this material.

    
*)

(* :Keywords:
	Parametric, Surface, Volume, Plot
*)

(* :Requirements: None. *)

(* :Warnings: None. *)

(* :Sources: 
       Andrei Constantinescu, Alexander Korsunsky - Classical Elasticity with Mathematica, Cambridge
*)

(* :Summary:
This package allows one to plot a 2D mesh in 2D or a 3D mesh in 3D and 
fill the mesh using a colorfunction depending of the coordinates of 
the nodes of the grid for each quadrangular or rectangular box.
It has a similar functionality to build-in Mathematica commands 
ParametricPlot, ParametricPlot3D
*)


BeginPackage["Graphics`ParametricMesh`"]

ParametricMesh::usage =
"ParametricMesh[{fx,fy}, {u,u0,u1,(du)},{v,v0,v1,(dv)} (options)]
plots a two-parameter mesh in a 2D space. fx and fy are functions of u and v.
 
ParametricMesh3D[{fx,fy,fz}, {u,u0,u1,(du)},{v,v0,v1,(dv),{w,w0,w1,(dw)} (options)]
plots a three-parameter mesh in a 3D space. 

Options are partly local: FillColor, Mesh, Fill, PlotPoints or passed to Show."

ParametricMesh3D::usage =
"ParametricMesh3D[{fu,fv,fw}, {u,u0,u1,(du)},{v,v0,v1,(dv),{w,w0,w1,(dw)} (options)]
plots a three-parameter mesh in a 3D space. fx, fy and fz are functions of 
u, v and w. 

ParametricMesh[{fu,fv}, {u,u0,u1,(du)},{v,v0,v1,(dv)} (options)]
plots a two-parameter mesh in a 2D space.

Options are partly local: FillColor, Mesh, Fill, PlotPoints or passed to Show."


ParametricBox3D::usage =
"ParametricMesh3D[{fu,fv,fw}, {u,u0,u1,(du)},{v,v0,v1,(dv),{w,w0,w1,(dw)} (options)]
plots a three-parameter box meshed on the surface in a 3D space. fx, fy and fz are functions of 
u, v and w. 

ParametricMesh[{fu,fv}, {u,u0,u1,(du)},{v,v0,v1,(dv)} (options)]
plots a two-parameter mesh in a 2D space.

Options are partly local: FillColor, Mesh, Fill, PlotPoints or passed to Show."


MeanNode::usage = "MeanNode[nodelist] = compute the mean node of the first 4 nodes of list of 5 nodes. 
                  We need 5 nodes to declare the closed Polygon" 


FillColor::usage = "FillColor is an Option for ParametricMesh and ParametricMesh3D defining the 
colorfunction for filling the elements of the mesh as a function of the positions of the vertices."
 
Fill::usage = "Fill is an Option for ParametricMesh and ParametricMesh3D defining the 
if the filling of the elements of the mesh is set to True or False."

Mesh::usage = "Mesh is an Option for ParametricMesh and ParametricMesh3D defining the 
if the mesh is going to be plotted True or False."

PlotPoints::usage = "Mesh is an Option for ParametricMesh and ParametricMesh3D defining the 
number of intermediate points for each parameter of the mesh."

ClosedLine::usage = "ClosedLine[listofpoint] is an extension of Line which closes the created line."

Begin["`Private`"]

Needs["Utilities`FilterOptions`"]

Options[ParametricMesh] = {PlotPoints -> 10, FillColor -> (Hue[ Norm[MeanNode[#]]] &) , Mesh -> True, Fill -> True}
Options[ParametricMesh3D] = {PlotPoints -> 10, FillColor -> (Hue[ Norm[MeanNode[#]]] &) , Mesh -> True, Fill -> True}
Options[ParametricBox3D] = {PlotPoints -> 10, FillColor -> (Hue[ Norm[MeanNode[#]]] &) , Mesh -> True, Fill -> True}

MeanNode[ nl_] := 0.25 Plus @@  Drop[nl, -1] 

ClosedLine[ ptlist_ ] := Line[ Append[ ptlist, First[ptlist]] ]

ParametricMesh[ {fu_, fv_}, {u_, u0_, u1_, du_:Automatic} ,{v_, v0_, v1_, dv_:Automatic}, opt___?OptionQ ] := 
   Module[{ndu = du, ndv = dv},
   
    fillcolor = FillColor /. {opt} /. Options[ParametricMesh]; 
    plotpoints = PlotPoints /. {opt} /. Options[ParametricMesh];
    mesh = Mesh /. {opt} /. Options[ParametricMesh];
    fill = Fill /. {opt} /. Options[ParametricMesh];
    
    If[ du === Automatic, ndu = N[(u1 - u0) / (plotpoints + 1) ] ];
    If[ dv === Automatic, ndv = N[(v1 - v0) / (plotpoints + 1) ] ]; 
    
    uu = Range @@ {u0, u1, ndu};
    vv = Range @@ {v0, v1, ndv};
    thenodes = Outer[ List, uu, vv] ;  
    meshnodes = Map[  {fu , fv} /.Thread[{u,v}\[Rule]#]&, thenodes, {2}];
    thequads = Map[   Flatten[#,1][[{1,2,4,3}]]  & ,Flatten[ Partition[ meshnodes, {2,2}, {1,1}], 1]];
    themesh = Map[ ClosedLine, thequads];
    thefill = Map[ {fillcolor[#] , Polygon[#]} &, thequads ];
    Which[ 
      mesh && fill, 
         Show[Graphics[thefill],Graphics[themesh ], FilterOptions[Graphics, opt] ], 
      mesh && Not[fill] , 
         Show[Graphics[themesh ], FilterOptions[Graphics, opt] ],
      Not[mesh] && fill, 
         Show[Graphics[thefill], FilterOptions[Graphics, opt] ] 
      ]
    
    ]

ParametricMesh3D[ {fu_, fv_, fz_}, {u_, u0_, u1_, du_:Automatic} ,{v_, v0_, v1_, dv_:Automatic}, opt___?OptionQ ] := 
   Module[{ndu = du, ndv = dv},
   
    fillcolor = FillColor /. {opt} /. Options[ParametricMesh]; 
    plotpoints = PlotPoints /. {opt} /. Options[ParametricMesh];
    mesh = Mesh /. {opt} /. Options[ParametricMesh];
    fill = Fill /. {opt} /. Options[ParametricMesh];
    
    If[ du === Automatic, ndu = N[(u1 - u0) / (plotpoints + 1) ] ];
    If[ dv === Automatic, ndv = N[(v1 - v0) / (plotpoints + 1) ] ]; 
    
    uu = Range @@ {u0, u1, ndu};
    vv = Range @@ {v0, v1, ndv};
    thenodes = Outer[ List, uu, vv] ;  
    meshnodes = Map[  {fu , fv, fz} /.Thread[{u,v}\[Rule]#]&, thenodes, {2}];
    thequads = Map[   Flatten[#,1][[{1,2,4,3}]]  & , Flatten[ Partition[ meshnodes, {2,2}, {1,1}], 1]];
    themesh = Map[ ClosedLine, thequads];
    thefill = Map[ {fillcolor[#] , Polygon[#]} &, thequads ];
    Which[ 
      mesh && fill, 
         Show[Graphics3D[thefill],Graphics3D[themesh ], FilterOptions[Graphics3D, opt] ], 
      mesh && Not[fill] , 
         Show[Graphics3D[themesh ], FilterOptions[Graphics3D, opt] ],
      Not[mesh] && fill, 
         Show[Graphics3D[thefill], FilterOptions[Graphics3D, opt] ] 
      ]
    
    ]


ParametricBox3D[ {fx_, fy_, fz_},  {u_, u0_, u1_, du_:Automatic} ,{v_, v0_, v1_, dv_:Automatic},{w_, w0_, w1_, dw_:Automatic}, opt___ ] :=
  Module[{ndu = du, ndv = dv, ndw = dw, puv, puw, pwv, ppopt},
    df = DisplayFunction -> Identity;

    If[ du === Automatic, ndu = N[(u1 - u0) / (plotpoints + 1) ] ];
    If[ dv === Automatic, ndv = N[(v1 - v0) / (plotpoints + 1) ] ]; 
    If[ dw === Automatic, ndw = N[(v1 - v0) / (plotpoints + 1) ] ]; 


    ppopt = FilterOptions[ ParametricMesh3D , {opt}];
    puv0 = ParametricMesh3D[ Evaluate[{fx,fy,fz} /. w -> w0 ] , {u, u0, u1, ndu} ,{v, v0, v1, ndv}, Evaluate[ppopt], Evaluate[df]];  
    puv1 = ParametricMesh3D[ Evaluate[{fx,fy,fz} /. w -> w1 ] , {u, u0, u1, ndu} ,{v, v0, v1, ndv}, Evaluate[ppopt], Evaluate[df]];  
    puw0 = ParametricMesh3D[ Evaluate[{fx,fy,fz} /. v -> v0 ] , {u, u0, u1, ndu} ,{w, w0, w1, ndw}, Evaluate[ppopt], Evaluate[df]];  
    puw1 = ParametricMesh3D[ Evaluate[{fx,fy,fz} /. v -> v1 ] , {u, u0, u1, ndu} ,{w, w0, w1, ndw}, Evaluate[ppopt], Evaluate[df]];  
    pwv0 = ParametricMesh3D[ Evaluate[{fx,fy,fz} /. u -> u0 ] , {w, w0, w1, ndw} ,{v, v0, v1, ndv}, Evaluate[ppopt], Evaluate[df]]; 
    pwv1 = ParametricMesh3D[ Evaluate[{fx,fy,fz} /. u -> u1 ] , {w, w0, w1, ndw} ,{v, v0, v1, ndv}, Evaluate[ppopt], Evaluate[df]]; 
    Show[ puv0, puw0, pwv0, puv1, puw1, pwv1 , FilterOptions[ Graphics3D, opt] , DisplayFunction -> $DisplayFunction]
  ]


End[]

Protect[ParametricMesh, ParametricMesh3D, ParametricBox3D, MeanNode, MeanNode3D]

EndPackage[]


(* :Examples:

ParametricMesh[{u + 2v ,v},{u, 0 , 0.7},{v,0,0.5}, PlotPoints -> 50 , Frame -> True, Mesh -> False, FillColor -> Function[ x, Hue[0.2]]]

ParametricMesh3D[{u + v, v , 0.7 + 0.1 ( u^2 + v^2)}, {u, 0, 3, 0.1}, {v, 0, 0.5, 0.1}, Lighting -> False]

ParametricBox3D[{u , v , w + 0.1 ( u^2 + v^2)}, {u, 0, 3, 0.2}, {v, 0, 0.5, 0.5},{w,-0.1, 0.1, 0.1}, Frame -> True, 
        Mesh -> False, FillColor -> (Hue[ Norm[MeanNode[#]]] &), ViewPoint -> {-1,-1,0.5}, Lighting -> False]

*)
